@@ -3,6 +3,7 @@
 
 #include "libfilezilla.hpp"
 
+#include <algorithm>
 #include <string>
 #include <vector>
 
@@ -118,6 +119,23 @@ String str_toupper_ascii(String const& s)
 	return ret;
 }
 
+/** \brief Comparator to be used for std::map for case-insentitive keys
+ *
+ * Comparision is done locale-agnostic.
+ * Useful for key-value pairs in protocols, e.g. HTTP headers.
+ */
+struct FZ_PUBLIC_SYMBOL less_insensitive_ascii final
+{
+	template<typename T>
+	bool operator()(T const& lhs, T const& rhs) const {
+		return std::lexicographical_compare(lhs.cbegin(), lhs.cend(), rhs.cbegin(), rhs.cend(),
+		    [](typename T::value_type const& a, typename T::value_type const& b) {
+			    return tolower_ascii(a) < tolower_ascii(b);
+		    }
+		);
+	}
+};
+
 /** \brief Converts from std::string in system encoding into std::wstring
  *
  * \return the converted string on success. On failure an empty string is returned.
@@ -142,6 +160,7 @@ inline typename std::enable_if<std::is_arithmetic<std::decay_t<Arg>>::value, std
  * \return the converted string on success. On failure an empty string is returned.
  */
 std::wstring FZ_PUBLIC_SYMBOL to_wstring_from_utf8(std::string const& in);
+std::wstring FZ_PUBLIC_SYMBOL to_wstring_from_utf8(char const* s, size_t len);
 
 /** \brief Converts from std::wstring into std::string in system encoding
  *
@@ -185,78 +204,6 @@ std::string FZ_PUBLIC_SYMBOL to_utf8(std::string const& in);
  */
 std::string FZ_PUBLIC_SYMBOL to_utf8(std::wstring const& in);
 
-
-/** \brief Converts a hex digit to decimal int
- *
- * Example: '9' becomes 9, 'b' becomes 11, 'D' becomes 13
- *
- * Returns -1 if input is not a valid hex digit.
- */
-template<typename Char>
-int hex_char_to_int(Char c)
-{
-	if (c >= 'a' && c <= 'z') {
-		return c - 'a' + 10;
-	}
-	if (c >= 'A' && c <= 'Z') {
-		return c - 'A' + 10;
-	}
-	else if (c >= '0' && c <= '9') {
-		return c - '0';
-	}
-	return -1;
-}
-
-template<typename String>
-std::vector<uint8_t> hex_decode(String const& in)
-{
-	std::vector<uint8_t> ret;
-	if (!(in.size() % 2)) {
-		ret.reserve(in.size() / 2);
-		for (size_t i = 0; i < in.size(); i += 2) {
-			int high = hex_char_to_int(in[i]);
-			int low = hex_char_to_int(in[i + 1]);
-			if (high == -1 || low == -1) {
-				return std::vector<uint8_t>();
-			}
-			ret.push_back(static_cast<uint8_t>((high << 4) + low));
-		}
-	}
-
-	return ret;
-}
-
-/** \brief Converts an integer to the corresponding lowercase hex digit
-*
-* Example: 9 becomes '9', 11 becomes 'b'
-*
-* Undefined output if input is less than 0 or larger than 15
-*/
-template<typename Char = char, bool Lowercase = true>
-Char int_to_hex_char(int d)
-{
-	if (d > 9) {
-		return (Lowercase ? 'a' : 'A') + d - 10;
-	}
-	else {
-		return '0' + d;
-	}
-}
-
-template<typename String, typename InString, bool Lowercase = true>
-String hex_encode(InString const& data)
-{
-	static_assert(sizeof(typename InString::value_type) == 1, "Input must be a container of 8 bit values");
-	String ret;
-	ret.reserve(data.size() * 2);
-	for (auto const& c : data) {
-		ret.push_back(int_to_hex_char<typename String::value_type, Lowercase>(static_cast<unsigned char>(c) >> 4));
-		ret.push_back(int_to_hex_char<typename String::value_type, Lowercase>(static_cast<unsigned char>(c) & 0xf));
-	}
-
-	return ret;
-}
-
 /// Calls either fz::to_string or fz::to_wstring depending on the passed template argument
 template<typename String, typename Arg>
 inline auto toString(Arg&& arg) -> typename std::enable_if<std::is_same<String, std::string>::value, decltype(to_string(std::forward<Arg>(arg)))>::type
@@ -328,11 +275,11 @@ Container strtok(String const& s, Delim const& delims)
 
 		// Not found, we're at ends;
 		if (pos == String::npos) {
-			if (start + 1 < s.size()) {
+			if (start < s.size()) {
 				ret.emplace_back(s.substr(start));
 			}
 		}
-		else if (pos > start + 1) {
+		else if (pos > start) {
 			// Non-empty substring
 			ret.emplace_back(s.substr(start, pos - start));
 		}
@@ -342,13 +289,6 @@ Container strtok(String const& s, Delim const& delims)
 	return ret;
 }
 
-
-/// \brief Encodes raw input string to base64
-std::string FZ_PUBLIC_SYMBOL base64_encode(std::string const& in);
-
-/// \brief Decodes base64, ignores whitespace. Returns empty string on invalid input.
-std::string FZ_PUBLIC_SYMBOL base64_decode(std::string const& in);
-
 // Converts string to integral type T. If string is not convertible, T() is returned.
 template<typename T, typename String>
 T to_integral(String const& s, T const errorval = T())
@@ -395,24 +335,63 @@ bool str_is_ascii(String const& s) {
 
 /// \brief Return passed string with all leading and trailing whitespace removed
 template<typename String>
-String trimmed(String const& s) {
-	size_t const first = s.find_first_not_of(fzS(typename String::value_type, " \r\n\t"));
+String trimmed(String const& s, String const& chars = fzS(typename String::value_type, " \r\n\t"), bool fromLeft = true, bool fromRight = true) {
+	size_t const first = fromLeft ? s.find_first_not_of(chars) : 0;
 	if (first == String::npos) {
 		return String();
 	}
-	else {
-		// Cannot be npos
-		size_t const last = s.find_last_not_of(fzS(typename String::value_type, " \r\n\t"));
 
-		return s.substr(first, last - first + 1);
+	size_t const last = fromRight ? s.find_last_not_of(chars) : s.size();
+	if (last == String::npos) {
+		return String();
 	}
+	return s.substr(first, last - first + 1);
 }
 
+template<typename String>
+String ltrimmed(String const& s, String const& chars = fzS(typename String::value_type, " \r\n\t")) {
+	return trimmed(s, chars, true, false);
+}
+
+template<typename String>
+String rtrimmed(String const& s, String const& chars = fzS(typename String::value_type, " \r\n\t")) {
+	return trimmed(s, chars, false, true);
+}
 
 /// \brief Remove all leading and trailing whitespace from string
 template<typename String>
-void trim(String & s) {
-	s = trimmed(s);
+void trim(String & s, String const& chars = fzS(typename String::value_type, " \r\n\t")) {
+	s = trimmed(s, chars);
+}
+
+template<typename String>
+void ltrim(String & s, String const& chars = fzS(typename String::value_type, " \r\n\t")) {
+	s = trimmed(s, chars, true, false);
+}
+
+template<typename String>
+void rtrim(String & s, String const& chars = fzS(typename String::value_type, " \r\n\t")) {
+	s = trimmed(s, chars, false, true);
+}
+
+// Remove once C++20 is minimum required standard to build libfilezilla
+template<typename String>
+bool starts_with(String const& s, String const& beginning)
+{
+	if (beginning.size() > s.size()) {
+		return false;
+	}
+	return std::equal(beginning.begin(), beginning.end(), s.begin());
+}
+
+// Remove once C++20 is minimum required standard to build libfilezilla
+template<typename String>
+bool ends_with(String const& s, String const& ending)
+{
+	if (ending.size() > s.size()) {
+		return false;
+	}
+	return std::equal(ending.rbegin(), ending.rend(), s.rbegin());
 }
 
 }
